"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.AWSEventBus = void 0;
const client_sts_1 = require("@aws-sdk/client-sts");
const client_sns_1 = require("@aws-sdk/client-sns");
const client_sqs_1 = require("@aws-sdk/client-sqs");
const graphql_eventbus_1 = require("graphql-eventbus");
const s3_1 = require("./s3");
const client_s3_1 = require("@aws-sdk/client-s3");
const uuid_1 = require("uuid");
const config_1 = require("./config");
class AWSEventBus {
    constructor(config) {
        var _a, _b;
        this.config = config;
        this.publishTopics = {};
        this.closeSignal = false;
        this.deleteSubscriptionsAndQueues = [];
        this.ongoingPublishes = new Set();
        this.receiveMessageFromQueue = async (queueUrl, cb) => {
            var _a, _b, _c, _d;
            try {
                // Receive messages from the queue
                const receiveMessageCommand = new client_sqs_1.ReceiveMessageCommand({
                    QueueUrl: queueUrl, // The URL of the SQS queue
                    MaxNumberOfMessages: (_b = (_a = this.config.subscriber) === null || _a === void 0 ? void 0 : _a.maxNumberOfMessages) !== null && _b !== void 0 ? _b : 10, // Number of messages to retrieve (max is 10)
                    WaitTimeSeconds: (_d = (_c = this.config.subscriber) === null || _c === void 0 ? void 0 : _c.pollingTimeSeconds) !== null && _d !== void 0 ? _d : 20, // Long polling (wait for messages up to 20 seconds)
                    VisibilityTimeout: 30, // The time for which a message is hidden after being received
                    AttributeNames: ["All"], // Optionally retrieve additional message attributes
                    MessageAttributeNames: ["All"], // Optionally retrieve all message attributes
                });
                // Send the command to receive the message
                const response = await this.sqsClient.send(receiveMessageCommand);
                if (this.closeSignal) {
                    return;
                }
                for (const message of response.Messages || []) {
                    const messageBody = JSON.parse(message.Body || "");
                    // Process the message
                    // Your custom logic for processing the message goes here
                    cb(JSON.parse(messageBody.Message)).then(() => {
                        const deleteMessageCommand = new client_sqs_1.DeleteMessageCommand({
                            QueueUrl: queueUrl,
                            ReceiptHandle: message.ReceiptHandle, // Required to delete the message
                        });
                        // Delete the message from the queue to avoid reprocessing it
                        this.sqsClient.send(deleteMessageCommand);
                    });
                }
            }
            catch (error) {
                if (this.closeSignal) {
                    return;
                }
                console.error("Error receiving or deleting message:", error);
            }
        };
        this.pollQueue = async (queueUrl, topicName, cb) => {
            const foo = () => this.receiveMessageFromQueue(queueUrl, async (baggage) => {
                if ("__s3Key" in baggage) {
                    const command = new client_s3_1.GetObjectCommand({
                        Bucket: config_1.Config.bucket,
                        Key: baggage["__s3Key"],
                    });
                    const response = await (0, s3_1.getS3)().send(command);
                    if (!response.Body) {
                        return;
                    }
                    try {
                        baggage = JSON.parse(await response.Body.transformToString("utf-8"));
                    }
                    catch (e) {
                        console.error(`Error in parsing the baggage payload for key ${baggage["__s3Key"]}`);
                        return;
                    }
                }
                await cb({
                    topic: topicName,
                    baggage: {
                        metadata: baggage.metadata,
                        payload: baggage.payload,
                    },
                });
            });
            while (!this.closeSignal) {
                await foo();
            }
            // this.pollTimers.push(setInterval(foo, 1000));
        };
        this.createTopic = async (topicName) => {
            var _a;
            if (!this.existingTopicsArns) {
                this.existingTopicsArns = await this.snsClient.send(new client_sns_1.ListTopicsCommand({}));
            }
            let topicArn;
            for (const topic of this.existingTopicsArns.Topics || []) {
                if ((_a = topic.TopicArn) === null || _a === void 0 ? void 0 : _a.endsWith(`:${topicName}`)) {
                    topicArn = topic.TopicArn;
                    console.log(`Topic already exists for: ${topicName}`);
                    break;
                }
            }
            if (!topicArn) {
                // eslint-disable-next-line no-await-in-loop
                const createTopicResponse = await this.snsClient.send(new client_sns_1.CreateTopicCommand({ Name: topicName }));
                topicArn = createTopicResponse.TopicArn;
            }
            if (!topicArn) {
                throw new Error(`Topic creation failed for topic ${topicArn}`);
            }
            return topicArn;
        };
        this.createQueue = async (queueName, topicArn) => {
            var _a;
            let queueUrl;
            try {
                // Step 1: Check if the queue already exists
                const getQueueUrlResponse = await this.sqsClient.send(new client_sqs_1.GetQueueUrlCommand({ QueueName: queueName }));
                queueUrl = getQueueUrlResponse.QueueUrl;
                console.log(`Queue already exists with URL: ${queueUrl}`);
            }
            catch (error) {
                if (error instanceof Error && error.name === "QueueDoesNotExist") {
                    console.log(`Queue does not exist. Creating queue: ${queueName}`);
                    const stsResponse = await this.stsClient.send(new client_sts_1.GetCallerIdentityCommand({}));
                    const accountId = stsResponse.Account;
                    const queueArn = `arn:aws:sqs:${this.config.region}:${accountId}:${queueName}`;
                    const policy = {
                        Version: "2012-10-17",
                        Statement: [
                            {
                                Effect: "Allow",
                                Principal: "*",
                                Action: "SQS:SendMessage",
                                Resource: queueArn,
                                Condition: {
                                    ArnEquals: {
                                        "aws:SourceArn": topicArn, // Restrict access to the specific SNS topic ARN
                                    },
                                },
                            },
                        ],
                    };
                    const createQueueResponse = await this.sqsClient.send(new client_sqs_1.CreateQueueCommand({
                        QueueName: queueName,
                        Attributes: {
                            Policy: JSON.stringify(policy),
                        },
                    }));
                    queueUrl = createQueueResponse.QueueUrl;
                    console.log(`Queue created with URL: ${queueUrl}`);
                }
                else {
                    console.error("Error finding or creating queue:", error);
                    throw error;
                }
            }
            if (!queueUrl) {
                throw new Error(`queue creation failed ${queueName}`);
            }
            const getQueueAttributesResponse = await this.sqsClient.send(new client_sqs_1.GetQueueAttributesCommand({
                QueueUrl: queueUrl,
                AttributeNames: ["QueueArn"], // Specify that you want the QueueArn attribute
            }));
            const queueArn = (_a = getQueueAttributesResponse.Attributes) === null || _a === void 0 ? void 0 : _a.QueueArn;
            if (!queueArn) {
                throw new Error(`queue ARN retrieval failed ${queueArn}`);
            }
            return {
                queueArn,
                queueUrl,
            };
        };
        this.closeConsumer = async () => {
            this.closeSignal = true;
            await Promise.all(this.deleteSubscriptionsAndQueues.map(async ({ queueUrl, subscriptionArn }) => {
                console.log(`deleting queue ${queueUrl} and subscription ${subscriptionArn}`);
                const unsubscribeCommand = new client_sns_1.UnsubscribeCommand({
                    SubscriptionArn: subscriptionArn,
                });
                await this.snsClient.send(unsubscribeCommand);
                const deleteQueueCommand = new client_sqs_1.DeleteQueueCommand({
                    QueueUrl: queueUrl,
                });
                await this.sqsClient.send(deleteQueueCommand);
            })).catch(console.error);
        };
        this.closePublisher = async () => {
            await Promise.all(this.ongoingPublishes);
        };
        this.init = () => {
            return this.bus.init();
        };
        this.publish = async (a) => {
            await this.bus.publish({
                payload: a.payload,
                topic: a.topic,
                metadata: a.metadata,
                extra: {
                    attributes: a.attributes,
                },
            });
        };
        this.snsClient = new client_sns_1.SNSClient({
            region: config.region,
        });
        this.sqsClient = new client_sqs_1.SQSClient({
            region: config.region,
        });
        this.stsClient = new client_sts_1.STSClient({ region: config.region });
        this.bus = new graphql_eventbus_1.GraphQLEventbus({
            plugins: config.plugins,
            publisher: this.config.publisher
                ? {
                    publishInit: async (topics) => {
                        for (const topicName of topics) {
                            // eslint-disable-next-line no-await-in-loop
                            this.publishTopics[topicName] = await this.createTopic(`graphql-eventbus-${topicName}`);
                        }
                    },
                    schema: (_a = this.config.publisher) === null || _a === void 0 ? void 0 : _a.schema,
                    publish: async (a) => {
                        let message = JSON.stringify(a.baggage);
                        const attributes = Object.entries(a.extra.attributes || {}).reduce((acc, [key, value]) => {
                            return {
                                ...acc,
                                [key]: {
                                    DataType: "String",
                                    StringValue: value,
                                },
                            };
                        }, 
                        // https://docs.aws.amazon.com/sns/latest/dg/attribute-key-matching.html
                        // we must have a non empty attribute for filtering to work for non existing key
                        {
                            dummy: {
                                DataType: "String",
                                StringValue: "dummy",
                            },
                        });
                        const maxSize = 256 * 1024; // 256 KB SNS message limit
                        const messageSize = Buffer.byteLength(message, "utf8");
                        if (messageSize > maxSize) {
                            console.log("large message detected...");
                            const localS3 = (0, s3_1.getS3)();
                            const keyPrefix = config_1.Config.s3Prefix;
                            const key = keyPrefix + "/" + (0, uuid_1.v4)() + ".json";
                            console.log("uploading payload to ", key);
                            const command = new client_s3_1.PutObjectCommand({
                                Bucket: config_1.Config.bucket,
                                Key: key,
                                Body: message,
                            });
                            await localS3.send(command);
                            console.log("payload uploaded");
                            message = JSON.stringify({
                                __s3Key: key,
                            });
                        }
                        const publishCommand = new client_sns_1.PublishCommand({
                            TopicArn: this.publishTopics[a.topic],
                            Message: message, // The message content (JSON)
                            MessageAttributes: attributes, // The attributes for the message,
                        });
                        const publishPromise = this.snsClient.send(publishCommand);
                        this.ongoingPublishes.add(publishPromise);
                        try {
                            await publishPromise;
                        }
                        catch (e) {
                            console.error(e);
                            throw e;
                        }
                        finally {
                            this.ongoingPublishes.delete(publishPromise);
                        }
                    },
                }
                : undefined,
            subscriber: this.config.subscriber
                ? {
                    cb: (_b = this.config.subscriber) === null || _b === void 0 ? void 0 : _b.cb,
                    subscribe: async (allTopics, cb) => {
                        var _a, _b, _c, _d;
                        let finalTopics = allTopics;
                        if ((_b = (_a = this.config.subscriber) === null || _a === void 0 ? void 0 : _a.includeTopics) === null || _b === void 0 ? void 0 : _b.length) {
                            finalTopics = allTopics.filter((t) => { var _a, _b; return (_b = (_a = this.config.subscriber) === null || _a === void 0 ? void 0 : _a.includeTopics) === null || _b === void 0 ? void 0 : _b.includes(t); });
                        }
                        if ((_d = (_c = this.config.subscriber) === null || _c === void 0 ? void 0 : _c.skipTopics) === null || _d === void 0 ? void 0 : _d.length) {
                            finalTopics = allTopics.filter((t) => { var _a, _b; return !((_b = (_a = this.config.subscriber) === null || _a === void 0 ? void 0 : _a.skipTopics) === null || _b === void 0 ? void 0 : _b.includes(t)); });
                        }
                        await finalTopics.reduce(async (acc, topicName) => {
                            var _a, _b, _c, _d, _e, _f, _g, _h;
                            const topicArn = await this.createTopic(`graphql-eventbus-${topicName}`);
                            let subscriptionName = `graphql-eventbus-${this.config.serviceName}-${topicName}${this.config.isDarkRelease ? "-dark" : ""}`;
                            if ((_a = this.config.subscriber) === null || _a === void 0 ? void 0 : _a.version) {
                                subscriptionName = `graphql-eventbus-${this.config.serviceName}-${topicName}-${(_b = this.config.subscriber) === null || _b === void 0 ? void 0 : _b.version}`;
                            }
                            const isFanout = (_d = (_c = this.config.subscriber) === null || _c === void 0 ? void 0 : _c.fanoutTopics) === null || _d === void 0 ? void 0 : _d.includes(topicName);
                            // we use a different subscription name for each instance of the service for a fanout topic
                            if (isFanout) {
                                subscriptionName = `graphql-eventbus-${this.config.serviceName}-${topicName}-${Math.random().toString().split(".")[1]}${this.config.isDarkRelease ? "-dark" : ""}`;
                                if ((_e = this.config.subscriber) === null || _e === void 0 ? void 0 : _e.version) {
                                    subscriptionName = `graphql-eventbus-${this.config.serviceName}-${topicName}-${Math.random().toString().split(".")[1]}-${(_f = this.config.subscriber) === null || _f === void 0 ? void 0 : _f.version}`;
                                }
                            }
                            const { queueArn, queueUrl } = await this.createQueue(subscriptionName, topicArn);
                            let filterPolicy = {
                                [`x-prop-${this.config.serviceName}-dark`]: [
                                    { exists: false },
                                ],
                                [`x-prop-${this.config.serviceName}-version`]: [
                                    { exists: false },
                                ],
                            };
                            if (this.config.isDarkRelease) {
                                filterPolicy = {
                                    [`x-prop-${this.config.serviceName}-dark`]: ["true"],
                                };
                            }
                            if ((_g = this.config.subscriber) === null || _g === void 0 ? void 0 : _g.version) {
                                filterPolicy = {
                                    [`x-prop-${this.config.serviceName}-version`]: [
                                        (_h = this.config.subscriber) === null || _h === void 0 ? void 0 : _h.version,
                                    ],
                                };
                            }
                            const subscribeCommand = new client_sns_1.SubscribeCommand({
                                TopicArn: topicArn,
                                Protocol: "sqs",
                                Endpoint: queueArn,
                                Attributes: {
                                    FilterPolicy: JSON.stringify(filterPolicy),
                                    FilterPolicyScope: "MessageAttributes",
                                },
                            });
                            const response = await this.snsClient.send(subscribeCommand);
                            if (isFanout) {
                                response.SubscriptionArn &&
                                    this.deleteSubscriptionsAndQueues.push({
                                        queueUrl: queueUrl,
                                        subscriptionArn: response.SubscriptionArn,
                                    });
                            }
                            this.pollQueue(queueUrl, topicName, cb);
                            return acc.then(() => Promise.resolve());
                        }, Promise.resolve());
                    },
                    queries: this.config.subscriber.queries,
                    schema: this.config.subscriber.schema,
                }
                : undefined,
        });
    }
}
exports.AWSEventBus = AWSEventBus;
//# sourceMappingURL=AWSEventBus.js.map